# type: ignore
"""
Prepare data for GTEx dataset.

Data needs to be sampled, otherwise it:
- Inflates GitHub repo size
- Inflates the size of binary and source distributions
- Increase the load times of the load_gtex function
- Increase times of tests that rely on this dataset

Sampling:
- Dataset is reduced to only genes from chr21 and first 30 samples
- GTF file is reduced to just chr21 entries. Additionally, redundant attributes
  are removed to save space

"""
import csv
from pathlib import Path

import pandas as pd

OUTPUT_GTF_PATH = Path(__file__).parent / "gencode.v26.annotation.chr21.gtf.gz"
OUTPUT_EXP_PATH = Path(__file__).parent / "gtex_lung.first30.chr21.csv.gz"


def prepare_gtf(input_gtf_path):
    gtf = pd.read_csv(input_gtf_path, sep="\t", comment="#", header=None)

    # Take only chr21
    gtf = gtf[gtf[0] == "chr21"]
    # Take only gene and exon lines, since only this is needed for union exon
    # length computation
    gtf = gtf[gtf[2].isin(["gene", "exon"])]
    # Remove redundant attributes

    def keep_gene_id_name(x):
        elements = x.split(";")
        elements = [e.strip() for e in elements]
        elements = [e for e in elements if e.startswith("gene_name") or e.startswith("gene_id")]
        out = "; ".join(elements) + ";"
        return out

    gtf[8] = gtf[8].apply(keep_gene_id_name)

    gtf.to_csv(
        OUTPUT_GTF_PATH,
        sep="\t",
        header=False,
        index=False,
        quoting=csv.QUOTE_NONE,
    )

    gene_ids = gtf[8].str.extract(r'gene_id "([^"]+)";', expand=False).unique()
    return gene_ids


def prepare_exp(input_exp_path, gene_ids):
    exp = pd.read_csv(input_exp_path, sep="\t", skiprows=2, index_col=1)
    exp.drop(columns=["Description", "id"], inplace=True)
    exp.index.name = "ENSEMBL_ID"
    # Sort to make it reproducible and cast to list
    common_gene_ids = sorted(set(gene_ids).intersection(exp.index))
    exp = exp.loc[common_gene_ids]
    exp = exp.iloc[:, :30]

    # Save the transposed version
    exp.T.to_csv(OUTPUT_EXP_PATH)


if __name__ == "__main__":
    # First prepare GTF, since we need gene IDs from chr21 to extract data
    # from expression matrix

    # Downloaded from here
    # https://www.gencodegenes.org/human/release_26.html
    # https://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_human/release_26/gencode.v26.annotation.gtf.gz
    input_gtf_path = "/Users/me/Downloads/gencode.v26.annotation.gtf.gz"
    gene_ids = prepare_gtf(input_gtf_path)

    # Downloaded from here
    # https://gtexportal.org/home/datasets
    # https://storage.googleapis.com/gtex_analysis_v8/rna_seq_data/gene_reads/gene_reads_2017-06-05_v8_lung.gct.gz
    input_exp_path = "/Users/me/Downloads/gene_reads_2017-06-05_v8_lung.gct.gz"
    prepare_exp(input_exp_path, gene_ids)



"""Command line interface."""

import functools
import io
import sys
from pathlib import Path
from typing import Any, Callable, Optional, Type, Union

import click
import pandas as pd

from rnanorm import CPM, CTF, CUF, FPKM, TMM, TPM, UQ

method_type = Type[Union[CPM, FPKM, TMM, TPM, UQ]]
file_type = Union[io.TextIOWrapper, Path]


class CLWrapper:
    """Input / output wrapper around normalization methods."""

    def __init__(self, method: method_type, **kwargs: Any) -> None:
        """Initialize class."""
        self.method = method
        self.kwargs = kwargs

    def handle(self, exp: file_type, out: file_type, force: bool) -> None:
        """Parse, transform and write out the results."""
        X = self.parse_exp(exp)

        # Turn ``--gene-lengths`` (CLI option, a file object) into
        # ``gene_lengths`` (TPM / FPKM method argument, a Series)
        if self.kwargs.get("gene_lengths", None) is not None:
            self.kwargs["gene_lengths"] = self.parse_gene_lengths(self.kwargs["gene_lengths"])

        normalization = self.method(**self.kwargs).set_output(transform="pandas")
        result = normalization.fit_transform(X)
        self.write_out(out, result, force)

    def parse_exp(self, exp: file_type) -> pd.DataFrame:
        """Parse expression matrix into a DataFrame."""
        if exp.name != "<stdin>":
            if not Path(exp.name).exists():
                raise ValueError(f"File {exp.name} does not exist.")
            if not Path(exp.name).is_file():
                raise ValueError(f"File {exp.name} is not a file.")

        return pd.read_csv(exp, index_col=0)

    def parse_gene_lengths(self, gene_lengths_file: file_type) -> pd.Series:
        """Parse gene lengths file into a Series object."""
        if not Path(gene_lengths_file.name).exists():
            raise ValueError(f"File {gene_lengths_file.name} does not exist.")
        if not Path(gene_lengths_file.name).is_file():
            raise ValueError(f"File {gene_lengths_file.name} is not a file.")

        return pd.read_csv(gene_lengths_file, index_col=0).iloc[:, 0]

    def write_out(self, out: file_type, result: pd.DataFrame, force: bool = False) -> None:
        """Write the resulting DataFrame to a file or stdout."""
        if out.name != "<stdout>":
            if Path(out.name).is_dir():
                raise ValueError(f"File {out.name} is a directory.")
            elif Path(out.name).exists() and not force:
                raise ValueError(f"File {out.name} already exists. Use --force flag to overwrite.")
            elif not Path(out.name).parent.exists():
                Path(out.name).parent.mkdir(parents=True)

        # result.index.name = "sample_id"
        result.to_csv(out)


def common_params(func: Callable[..., Any]) -> Callable[..., Any]:
    """Set common parameters for all normalization methods."""

    @click.argument("exp", type=click.File("r"), default=sys.stdin)
    @click.option(
        "--out",
        type=click.File("w"),
        default=sys.stdout,
        help="Output results in this file instead of stdout",
    )
    @click.option("--force", is_flag=True, help="Overwrite already existing output file")
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Callable[..., Any]:
        return func(*args, **kwargs)

    return wrapper


def gtf_param(func: Callable[..., Any]) -> Callable[..., Any]:
    """Set parameters for normalization methods that require a GTF file."""

    @click.option(
        "--gtf",
        type=click.File("r"),
        required=False,
        help="Compute gene lengths from this GTF file",
    )
    @click.option(
        "--gene-lengths",
        type=click.File("r"),
        required=False,
        help="File with gene lengths",
    )
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Callable[..., Any]:
        return func(*args, **kwargs)

    return wrapper


def tmm_params(func: Callable[..., Any]) -> Callable[..., Any]:
    """Set parameters for TMM and CTF normalization."""

    @click.option("--m_trim", default=0.3, help="Two sided cutoff for M-values")
    @click.option("--a_trim", default=0.05, help="Two sided cutoff for A-values")
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Callable[..., Any]:
        return func(*args, **kwargs)

    return wrapper


@click.command(short_help="Counts per million")
@common_params
def cpm(exp: pd.DataFrame, out: file_type, force: bool) -> None:
    """Compute CPM."""
    CLWrapper(CPM).handle(exp, out, force)


@click.command(short_help="Fragments per kilo-base million")
@common_params
@gtf_param
def fpkm(
    exp: pd.DataFrame,
    out: file_type,
    force: bool,
    gtf: Optional[io.TextIOWrapper] = None,
    gene_lengths: Optional[io.TextIOWrapper] = None,
) -> None:
    """Compute FPKM."""
    CLWrapper(FPKM, gtf=gtf, gene_lengths=gene_lengths).handle(exp, out, force)


@click.command(short_help="Transcripts per million")
@common_params
@gtf_param
def tpm(
    exp: pd.DataFrame,
    out: file_type,
    force: bool,
    gtf: Optional[io.TextIOWrapper] = None,
    gene_lengths: Optional[io.TextIOWrapper] = None,
) -> None:
    """Compute TPM."""
    CLWrapper(TPM, gtf=gtf, gene_lengths=gene_lengths).handle(exp, out, force)


@click.command(short_help="Upper quartile")
@common_params
def uq(exp: pd.DataFrame, out: file_type, force: bool) -> None:
    """Compute UQ."""
    CLWrapper(UQ).handle(exp, out, force)


@click.command(short_help="Counts adjusted with UQ factors")
@common_params
def cuf(exp: pd.DataFrame, out: file_type, force: bool) -> None:
    """Compute CUF."""
    CLWrapper(CUF).handle(exp, out, force)


@click.command(short_help="Trimmed mean of M-values")
@common_params
@tmm_params
def tmm(exp: pd.DataFrame, out: file_type, force: bool, m_trim: float, a_trim: float) -> None:
    """Compute TMM."""
    CLWrapper(TMM, m_trim=m_trim, a_trim=a_trim).handle(exp, out, force)


@click.command(short_help="Counts adjusted with TMM factors")
@common_params
@tmm_params
def ctf(exp: pd.DataFrame, out: file_type, force: bool, m_trim: float, a_trim: float) -> None:
    """Compute CTF."""
    CLWrapper(CTF, m_trim=m_trim, a_trim=a_trim).handle(exp, out, force)


@click.group()
def main() -> None:
    """Common RNA-seq normalization methods."""
    pass


# Add sub-commands to the main command
main.add_command(cpm)
main.add_command(fpkm)
main.add_command(tpm)
main.add_command(uq)
main.add_command(cuf)
main.add_command(tmm)
main.add_command(ctf)



"""
Configuration file for the Sphinx documentation builder.

For the full list of built-in configuration values, see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""
from importlib.metadata import metadata

meta = metadata("rnanorm")

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "RNA-seq normalization"
author = meta["Author"]
release = meta["Version"]
copyright = "2023, " + author

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx_autodoc_typehints",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
]

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# intersphinx mapping configuration
intersphinx_mapping = {
    "matplotlib": ("https://matplotlib.org/stable", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "python": ("https://docs.python.org/3", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
}



"""Commonly used type hints."""

import typing

import numpy as np
import numpy.typing as npt
from pandera.typing import DataFrame, Series

#: Type for a 1D matrix of numeric values.
Numeric1D = typing.Union[Series[float], npt.NDArray[np.float64]]
#: Type for a 2D matrix of numeric values.
Numeric2D = typing.Union[DataFrame[float], npt.NDArray[np.float64]]

if hasattr(typing, "Self"):
    Self = typing.Self
else:
    Self = object



"""Within sample normalization: CPM, FPKM and TPM."""

import warnings
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd
from sklearn.preprocessing import FunctionTransformer
from sklearn.utils.validation import validate_data

from ..annotation import GTF
from ..typing import Numeric1D, Numeric2D
from .utils import LibrarySize


class CPM(FunctionTransformer):
    """Counts per million (CPM) normalization.

    :param allow_nan: If true, allow X to contain NaN values.

    .. rubric:: Examples

    >>> from rnanorm.datasets import load_toy_data
    >>> from rnanorm import CPM
    >>> X = load_toy_data().exp
    >>> X
              Gene_1  Gene_2  Gene_3  Gene_4  Gene_5
    Sample_1     200     300     500    2000    7000
    Sample_2     400     600    1000    4000   14000
    Sample_3     200     300     500    2000   17000
    Sample_4     200     300     500    2000    2000
    >>> CPM().set_output(transform="pandas").fit_transform(X)
               Gene_1   Gene_2    Gene_3    Gene_4    Gene_5
    Sample_1  20000.0  30000.0   50000.0  200000.0  700000.0
    Sample_2  20000.0  30000.0   50000.0  200000.0  700000.0
    Sample_3  10000.0  15000.0   25000.0  100000.0  850000.0
    Sample_4  40000.0  60000.0  100000.0  400000.0  400000.0

    """

    def __init__(self, allow_nan: bool = False) -> None:
        """Initialize class."""
        self.allow_nan = allow_nan
        super().__init__(
            self._func,
            validate=True,
            feature_names_out="one-to-one",
        )

    def _func(self, X: Numeric2D) -> Numeric2D:
        """Transform.

        :param X: Expression raw count matrix (n_samples, n_features)
        :return: Normalized expression matrix (n_samples, n_features)
        """
        lib_size = LibrarySize(allow_nan=self.allow_nan).fit_transform(X)

        if hasattr(lib_size, "columns"):
            # In the case
            #   - .set_output(transform="pandas")
            #   - X is instance of pd.DataFrame
            # lib_size will be pd.DataFrame object with single
            # "Library size" column and below statement will fail with
            # InvalidIndexError. To avoid this - cast to numpy.
            lib_size = np.asarray(lib_size["Library size"])

        return X / lib_size[:, np.newaxis] * 1e6

    def _check_input(self, X: Numeric2D, *, reset: bool) -> Numeric2D:
        """Check input.

        This function is modified only so that NaN values can be
        tolerated even if self.validate=True.
        """
        kwargs = dict()
        if self.allow_nan:
            kwargs["ensure_all_finite"] = "allow-nan"
        # return self._validate_data(X, reset=reset, **kwargs)
        return validate_data(self, X, reset=reset, **kwargs)


class BaseNormalizationWithGTF(FunctionTransformer):
    """Normalize raw counts."""

    def __init__(
        self,
        gtf: Optional[Union[str, Path]] = None,
        gene_lengths: Optional[pd.Series] = None,
    ) -> None:
        """Initialize class."""
        if (gtf is None and gene_lengths is None) or (
            gtf is not None and gene_lengths is not None
        ):
            raise ValueError("One and only one of gtf or gene_lengths should be provided")
        self.gene_lengths = gene_lengths
        self.gtf = gtf
        super().__init__(
            func=self._func,
            validate=True,
            feature_names_out="one-to-one",
        )

    def _get_gene_lengths(self, X: Numeric2D) -> Numeric1D:
        """Check and correct X and gene lengths."""
        # gene_lengths should be a pandas.Series object with gene ID's in index
        # and gene lengths in values.
        if self.gtf:
            if not hasattr(self, "feature_names_in_"):
                raise ValueError("X should be a pandas.DataFrame object when using GTF file")
            gene_lengths = GTF(self.gtf).length
        else:
            # Validate gene lengths
            if not isinstance(self.gene_lengths, (pd.Series, np.ndarray)):
                raise ValueError("gene_lengths should be a pandas.Series or numpy.ndarray object")
            if not (self.gene_lengths > 0).all():
                raise ValueError("gene_lengths should only contain positive numbers")
            if not (self.gene_lengths - self.gene_lengths.astype(int)).sum() == 0:
                raise ValueError("gene_lengths should only contain integers")
            if isinstance(self.gene_lengths, pd.Series) and not self.gene_lengths.index.is_unique:
                raise ValueError("gene_lengths should have unique index")
            if (
                isinstance(self.gene_lengths, np.ndarray)
                and self.gene_lengths.shape[0] != X.shape[1]
            ):
                raise ValueError(
                    "gene_lengths should have same length as number of features in X."
                )

            gene_lengths = self.gene_lengths

        if hasattr(self, "feature_names_in_"):
            missing = set(self.feature_names_in_) - set(gene_lengths.index)
            if missing:
                if len(missing) == len(self.feature_names_in_):
                    container = "gtf" if self.gtf is not None else "gene_lengths"
                    raise ValueError(f"None of the genes in X are in {container}.")
                warnings.warn(
                    f"X contains {len(missing)} genes that are not in GTF "
                    f"{self.gtf}. This will result in NaN values for missing "
                    "genes in the output."
                )
            gene_lengths = gene_lengths.reindex(index=self.feature_names_in_).to_numpy()

        return gene_lengths


class FPKM(BaseNormalizationWithGTF):
    """Fragments per kilo-base million (FKPM) normalization.

    :param gtf: GTF file used when assigning counts to each gene

    .. rubric:: Examples

    >>> from rnanorm.datasets import load_toy_data
    >>> from rnanorm import FPKM
    >>> dataset = load_toy_data()
    >>> dataset.exp
              Gene_1  Gene_2  Gene_3  Gene_4  Gene_5
    Sample_1     200     300     500    2000    7000
    Sample_2     400     600    1000    4000   14000
    Sample_3     200     300     500    2000   17000
    Sample_4     200     300     500    2000    2000
    >>> fpkm = FPKM(gtf=dataset.gtf_path).set_output(transform="pandas")
    >>> fpkm.fit_transform(dataset.exp)
                Gene_1    Gene_2    Gene_3    Gene_4    Gene_5
    Sample_1  100000.0  100000.0  100000.0  200000.0  700000.0
    Sample_2  100000.0  100000.0  100000.0  200000.0  700000.0
    Sample_3   50000.0   50000.0   50000.0  100000.0  850000.0
    Sample_4  200000.0  200000.0  200000.0  400000.0  400000.0
    """

    def _func(self, X: Numeric2D) -> Numeric2D:
        gene_lengths = self._get_gene_lengths(X)
        return CPM().fit_transform(X) / gene_lengths * 1e3


class TPM(BaseNormalizationWithGTF):
    """Transcripts per kilo-base million (TPM) normalization.

    .. rubric:: Examples

    >>> from rnanorm.datasets import load_toy_data
    >>> from rnanorm import TPM
    >>> dataset = load_toy_data()
    >>> dataset.exp
              Gene_1  Gene_2  Gene_3  Gene_4  Gene_5
    Sample_1     200     300     500    2000    7000
    Sample_2     400     600    1000    4000   14000
    Sample_3     200     300     500    2000   17000
    Sample_4     200     300     500    2000    2000
    >>> tpm = TPM(gtf=dataset.gtf_path).set_output(transform="pandas")
    >>> tpm.fit_transform(dataset.exp)
                 Gene_1     Gene_2     Gene_3     Gene_4     Gene_5
    Sample_1   83333.33   83333.33   83333.33  166666.66  583333.33
    Sample_2   83333.33   83333.33   83333.33  166666.66  583333.33
    Sample_3   45454.54   45454.54   45454.54   90909.09  772727.27
    Sample_4  142857.14  142857.14  142857.14  285714.28  285714.28

    """

    def _func(self, X: Numeric2D) -> Numeric2D:
        gene_lengths = self._get_gene_lengths(X)
        # Exceptionally NaN values are allowed in case of missing genes:
        return CPM(allow_nan=True).fit_transform(X / gene_lengths * 1e3)



import pandas as pd

from rnanorm.annotation import GTF


def test_union_exon_lengths(gtf_file, exp):
    gene_lengths = GTF(gtf_file).length
    assert len(gene_lengths) == 5
    pd.testing.assert_series_equal(
        gene_lengths,
        pd.Series(
            [200, 300, 500, 1000, 1000],
            index=pd.Series(["Gene_1", "Gene_2", "Gene_3", "Gene_4", "Gene_5"], name="gene_id"),
            name="gene_length",
        ),
    )



"""Utilities for normalization."""

import numpy as np
from sklearn.preprocessing import FunctionTransformer
from sklearn.utils.validation import validate_data

from ..typing import Numeric2D


class LibrarySize(FunctionTransformer):
    """Library size.

    :param allow_nan: If true, allow X to contain NaN values.

    .. rubric:: Examples

    >>> from rnanorm.datasets import load_toy_data
    >>> from rnanorm import LibrarySize
    >>> X = load_toy_data().exp
    >>> X
              Gene_1  Gene_2  Gene_3  Gene_4  Gene_5
    Sample_1     200     300     500    2000    7000
    Sample_2     400     600    1000    4000   14000
    Sample_3     200     300     500    2000   17000
    Sample_4     200     300     500    2000    2000
    >>> LibrarySize().set_output(transform="pandas").fit_transform(X)
              Library size
    Sample_1       10000.0
    Sample_2       20000.0
    Sample_3       20000.0
    Sample_4        5000.0

    """

    def __init__(self, allow_nan: bool = False) -> None:
        """Initialize class."""
        self.allow_nan = allow_nan
        super().__init__(
            func=np.nansum,
            kw_args=dict(axis=1),
            validate=True,
            feature_names_out=lambda self, input_features: ["Library size"],
        )

    def _check_input(self, X: Numeric2D, *, reset: bool) -> Numeric2D:
        """Check input.

        This function is modified only so that NaN values can be
        tolerated even if self.validate=True.
        """
        kwargs = dict()
        if self.allow_nan:
            kwargs["ensure_all_finite"] = "allow-nan"
        return validate_data(self, X, reset=reset, **kwargs)



import numpy as np
import pandas as pd
import pytest

from rnanorm import CTF


@pytest.fixture
def expected_ctf(exp):
    return pd.DataFrame(
        [
            [200, 300, 500, 2000, 7000],
            [400, 600, 1000, 4000, 14000],
            [400, 600, 1000, 4000, 34000],
            [100, 150, 250, 1000, 1000],
        ],
        index=exp.index,
        columns=exp.columns,
        dtype=np.float64,
    )


def test_ctf(exp, expected_factors, expected_ctf):
    """Test UQ normalization."""
    # Simple case: Fit some data, transform same data.
    transformer = CTF()
    transformer.fit(exp)
    factors = transformer.get_norm_factors(exp)
    transformed_data = transformer.transform(exp)

    np.testing.assert_allclose(factors, expected_factors, rtol=1e-4)
    np.testing.assert_allclose(transformed_data, expected_ctf, rtol=1e-4)

    # Advanced case: Fit some samples, transform different ones.
    samples_to_fit = ["Sample_1", "Sample_3", "Sample_4"]
    exp1 = exp.loc[samples_to_fit]
    transformer.fit(exp1)
    factors = transformer.get_norm_factors(exp1)
    transformed_data = transformer.transform(exp.loc[["Sample_2"]])

    np.testing.assert_allclose(factors, [1, 0.5, 2.0], rtol=1e-4)
    np.testing.assert_allclose(
        transformed_data,
        expected_ctf.loc[["Sample_2"]],
        rtol=1e-3,
    )



import pandas as pd

from rnanorm import TPM
from rnanorm.datasets import load_gtex, load_toy_data


def test_toy_data():
    """Test that datasets.load_toy_data works as expected."""
    ds = load_toy_data()
    assert isinstance(ds.exp, pd.DataFrame)
    assert ds.exp.shape == (4, 5)
    assert ds.exp.index[0] == "Sample_1"
    assert ds.exp.columns[0] == "Gene_1"

    result = TPM(ds.gtf_path).set_output(transform="pandas").fit_transform(ds.exp)

    assert isinstance(result, pd.DataFrame)
    assert result.shape == (4, 5)


def test_gtex():
    """Test that datasets.load_gtex works as expected."""
    ds = load_gtex()
    assert isinstance(ds.exp, pd.DataFrame)
    assert ds.exp.shape == (30, 818)
    assert ds.exp.index[0] == "GTEX-111CU-0326-SM-5GZXO"
    assert ds.exp.columns[0] == "ENSG00000141956.13"

    result = TPM(ds.gtf_path).set_output(transform="pandas").fit_transform(ds.exp)

    assert isinstance(result, pd.DataFrame)
    assert result.shape == (30, 818)



"""Datasets."""

from pathlib import Path

import pandas as pd
from sklearn.utils import Bunch

FILES_DIR = Path(__file__).parent / "files"


def load_toy_data() -> Bunch:
    """
    Load a minimal, made-up raw count RNA-seq dataset.

    This dataset is not representative of any real RNA-seq dataset. However,
    it is small and has just enough complexity to showcase the effects of
    methods. It is also simple enough to make calculation by hand possible.

    .. rubric:: Examples

    >>> from rnanorm.datasets import load_toy_data
    >>> dataset = load_toy_data()
    >>> dataset.exp
              Gene_1  Gene_2  Gene_3  Gene_4  Gene_5
    Sample_1     200     300     500    2000    7000
    Sample_2     400     600    1000    4000   14000
    Sample_3     200     300     500    2000   17000
    Sample_4     200     300     500    2000    2000
    >>> # TPM and FPKM normalization also require GTF file
    >>> dataset.gtf_path
    PosixPath('/Users/me/.../toy.gtf')

    """
    ds = Bunch()
    ds.exp = pd.read_csv(FILES_DIR / "toy_exp.csv", index_col=0)
    ds.gtf_path = FILES_DIR / "toy.gtf"

    return ds


def load_gtex() -> Bunch:
    """
    Load a real RNA-seq dataset from GTFx project.

    Dataset is sampled to contain just chr21 and first 30 samples from GTEx
    lung V8.

    .. rubric:: Examples

    >>> from rnanorm.datasets import load_gtex
    >>> dataset = load_gtex()
    >>> dataset.exp
                              ENSG00000141956.13  ENSG00000141959.16   ...
    GTEX-111CU-0326-SM-5GZXO                 871                8129   ...
    GTEX-111FC-1126-SM-5GZWU                 852                7076   ...
    GTEX-111VG-0726-SM-5GIDC                 912               11016   ...
    ...                                      ...                 ...   ...
    >>> # TPM and FPKM normalization also require GTF file
    >>> dataset.gtf_path
    PosixPath('/Users/.../gencode.v26.annotation.chr21.gtf.gz')

    """
    ds = Bunch()
    ds.exp = pd.read_csv(FILES_DIR / "gtex_lung.first30.chr21.csv.gz", index_col=0)
    ds.gtf_path = FILES_DIR / "gencode.v26.annotation.chr21.gtf.gz"

    return ds



import numpy as np
import pandas as pd
import pytest

from rnanorm import LibrarySize


def test_lib_size():
    exp = pd.DataFrame(
        [
            [5, 5, 0],
            [0, 0, 0],
            [4, 1, 0],
        ],
        index=["SAMPLE_1", "SAMPLE_2", "SAMPLE_3"],
        columns=["GENE_1", "GENE_2", "GENE_3"],
        dtype=float,
    )
    expected = pd.DataFrame(
        [10.0, 0.0, 5.0],
        index=exp.index,
        columns=["Library size"],
    )

    transformer = LibrarySize()

    # Default output is np.ndarray
    lib_size = transformer.fit_transform(exp)
    assert isinstance(lib_size, np.ndarray)
    np.testing.assert_array_equal(lib_size, expected.to_numpy()[:, 0])

    # Check it possible to also get pandas output
    transformer.set_output(transform="pandas")
    lib_size = transformer.fit_transform(exp)
    assert isinstance(lib_size, pd.DataFrame)
    pd.testing.assert_frame_equal(lib_size, expected)

    # Validation refuses nan values
    exp.iloc[0, 0] = np.nan
    with pytest.raises(ValueError, match="Input X contains NaN."):
        transformer.fit_transform(exp)



import os
import subprocess
from io import StringIO

import pandas as pd
import pytest

from rnanorm import CPM, CTF, CUF, FPKM, TMM, TPM, UQ
from rnanorm.annotation import GTF


@pytest.fixture
def exp_path(tmp_path, exp):
    exp_path = tmp_path / "exp.csv"
    exp.to_csv(exp_path)
    return exp_path


@pytest.fixture
def out_path(tmp_path):
    out_path = tmp_path / "out.csv"
    yield out_path
    # Remove file after test
    out_path.unlink(missing_ok=True)


@pytest.fixture
def gene_lengths_file(tmp_path, gtf_file):
    out_path = tmp_path / "gene_lengths.csv"
    gene_lengths = GTF(gtf_file).length
    gene_lengths.to_csv(out_path)
    return out_path


def test_files_in_out(exp_path, out_path):
    assert not out_path.is_file()
    cp = subprocess.run(["rnanorm", "cpm", str(exp_path), "--out", str(out_path)])
    assert out_path.is_file()
    assert cp.returncode == 0


def test_stdout(exp_path, exp):
    """Parameter --out is not given, deliver via standard output."""
    cp = subprocess.run(
        ["rnanorm", "cpm", str(exp_path)],
        capture_output=True,
        text=True,
    )
    stream = StringIO(cp.stdout)
    stream.seek(0)
    pd.testing.assert_frame_equal(
        pd.read_csv(stream, index_col=0),
        CPM().set_output(transform="pandas").fit_transform(exp),
    )


def test_stdin(exp, out_path):
    """Input is not given, get input via standard input."""
    stream = StringIO()
    exp.to_csv(stream)
    stream.seek(0)

    cp = subprocess.run(
        ["rnanorm", "cpm", "--out", str(out_path)],
        input=stream.read(),
        text=True,
    )
    assert cp.returncode == 0
    pd.testing.assert_frame_equal(
        pd.read_csv(out_path, index_col=0),
        CPM().set_output(transform="pandas").fit_transform(exp),
    )


def test_fpkm(exp, exp_path, gtf_file, out_path, gene_lengths_file):
    assert not out_path.is_file()

    cp = subprocess.run(
        ["rnanorm", "fpkm", str(exp_path), "--out", str(out_path), "--gtf", str(gtf_file)]
    )
    assert out_path.is_file()
    assert cp.returncode == 0
    pd.testing.assert_frame_equal(
        pd.read_csv(out_path, index_col=0),
        FPKM(gtf=gtf_file).set_output(transform="pandas").fit_transform(exp),
    )

    # Test also with gene lengths file
    os.remove(out_path)
    assert not out_path.is_file()

    cp = subprocess.run(
        [
            "rnanorm",
            "fpkm",
            str(exp_path),
            "--out",
            str(out_path),
            "--gene-lengths",
            str(gene_lengths_file),
        ]
    )
    assert out_path.is_file()
    assert cp.returncode == 0
    pd.testing.assert_frame_equal(
        pd.read_csv(out_path, index_col=0),
        FPKM(gtf=gtf_file).set_output(transform="pandas").fit_transform(exp),
    )


def test_tpm(exp, exp_path, gtf_file, out_path, gene_lengths_file):
    assert not out_path.is_file()

    cp = subprocess.run(
        ["rnanorm", "tpm", str(exp_path), "--out", str(out_path), "--gtf", str(gtf_file)]
    )
    assert out_path.is_file()
    assert cp.returncode == 0
    pd.testing.assert_frame_equal(
        pd.read_csv(out_path, index_col=0),
        TPM(gtf=gtf_file).set_output(transform="pandas").fit_transform(exp),
    )

    # Test also with gene lengths file
    os.remove(out_path)
    assert not out_path.is_file()

    cp = subprocess.run(
        [
            "rnanorm",
            "tpm",
            str(exp_path),
            "--out",
            str(out_path),
            "--gene-lengths",
            str(gene_lengths_file),
        ]
    )
    assert out_path.is_file()
    assert cp.returncode == 0
    pd.testing.assert_frame_equal(
        pd.read_csv(out_path, index_col=0),
        TPM(gtf=gtf_file).set_output(transform="pandas").fit_transform(exp),
    )


def test_uq(exp, exp_path, out_path):
    assert not out_path.is_file()

    cp = subprocess.run(["rnanorm", "uq", str(exp_path), "--out", str(out_path)])
    assert out_path.is_file()
    assert cp.returncode == 0
    pd.testing.assert_frame_equal(
        pd.read_csv(out_path, index_col=0),
        UQ().set_output(transform="pandas").fit_transform(exp),
    )


def test_cuf(exp, exp_path, out_path):
    assert not out_path.is_file()

    cp = subprocess.run(["rnanorm", "cuf", str(exp_path), "--out", str(out_path)])
    assert out_path.is_file()
    assert cp.returncode == 0
    pd.testing.assert_frame_equal(
        pd.read_csv(out_path, index_col=0),
        CUF().set_output(transform="pandas").fit_transform(exp),
    )


def test_tmm(exp, exp_path, out_path):
    assert not out_path.is_file()

    cp = subprocess.run(["rnanorm", "tmm", str(exp_path), "--out", str(out_path)])
    assert out_path.is_file()
    assert cp.returncode == 0
    pd.testing.assert_frame_equal(
        pd.read_csv(out_path, index_col=0),
        TMM().set_output(transform="pandas").fit_transform(exp),
    )


def test_ctf(exp, exp_path, out_path):
    assert not out_path.is_file()

    cp = subprocess.run(["rnanorm", "ctf", str(exp_path), "--out", str(out_path)])
    assert out_path.is_file()
    assert cp.returncode == 0
    pd.testing.assert_frame_equal(
        pd.read_csv(out_path, index_col=0),
        CTF().set_output(transform="pandas").fit_transform(exp),
    )



import numpy as np
import pandas as pd
import pytest

from rnanorm import FPKM
from rnanorm.annotation import GTF


@pytest.fixture
def expected(exp):
    return pd.DataFrame(
        [
            [100000, 100000, 100000, 200000, 700000],
            [100000, 100000, 100000, 200000, 700000],
            [50000, 50000, 50000, 100000, 850000],
            [200000, 200000, 200000, 400000, 400000],
        ],
        columns=exp.columns,
        index=exp.index,
        dtype=np.float64,
    )


def test_fpkm(exp, expected, gtf_file):
    transformer = FPKM(gtf=gtf_file)

    # Numpy array is not allowed when using GTF
    with pytest.raises(ValueError, match=r"X should be a pandas.DataFrame"):
        exp_normalized = transformer.fit_transform(exp.values)

    # Default output is np.ndarray
    exp_normalized = transformer.fit_transform(exp)
    assert isinstance(exp_normalized, np.ndarray)
    np.testing.assert_array_equal(exp_normalized, expected.to_numpy())

    # Set pandas output
    transformer.set_output(transform="pandas")
    exp_normalized = transformer.fit_transform(exp)
    assert isinstance(exp_normalized, pd.DataFrame)
    pd.testing.assert_frame_equal(exp_normalized, expected)

    # Validation refuses nan values
    exp.iloc[0, 0] = np.nan
    with pytest.raises(ValueError, match="Input X contains NaN."):
        transformer.fit_transform(exp)


def test_fpkm_gene_lengths(exp, expected, gtf_file):
    gene_lengths = GTF(gtf=gtf_file).length

    # Numpy arrays are allowed when using gene lengths.
    # But numpy array must be subset ti have the same number of
    # genes as the gene lengths.
    transformer = FPKM(gene_lengths=gene_lengths.to_numpy())
    exp_normalized = transformer.fit_transform(exp.values)
    np.testing.assert_array_equal(exp_normalized, expected.to_numpy())

    # Work with pandas.
    transformer = FPKM(gene_lengths=gene_lengths)
    transformer.set_output(transform="pandas")
    exp_normalized = transformer.fit_transform(exp)
    assert isinstance(exp_normalized, pd.DataFrame)
    pd.testing.assert_frame_equal(exp_normalized, expected)



import numpy as np
import pandas as pd
from sklearn import config_context
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from rnanorm import CPM, CTF, CUF, FPKM, TMM, TPM, UQ
from rnanorm.datasets import load_toy_data


def test_grid_search():
    """Test compatibility of all methods with sklearn machinery."""
    ds = load_toy_data()
    X = ds.exp
    y = pd.Series([0, 0, 1, 1], index=X.index)
    pipeline = Pipeline(
        steps=[
            ("normalization", CPM()),
            ("scaler", StandardScaler()),
            ("classifier", LogisticRegression()),
        ]
    )
    params = {
        "normalization": [
            CPM(),
            FPKM(gtf=ds.gtf_path),
            TPM(gtf=ds.gtf_path),
            UQ(),
            CUF(),
            TMM(),
            CTF(),
        ],
    }
    search = GridSearchCV(pipeline, params, cv=2, refit=False)
    search.fit(X, y)
    results = pd.DataFrame(search.cv_results_)
    assert results.shape[0] == 7


def test_set_output():
    """Ensure set_output behaves as expected."""
    ds = load_toy_data()

    method__has_factors = (
        (CPM(), False),
        (FPKM(gtf=ds.gtf_path), False),
        (TPM(gtf=ds.gtf_path), False),
        (UQ(), True),
        (CUF(), True),
        (TMM(), True),
        (CTF(), True),
    )

    for method, has_factors in method__has_factors:
        # No configuration returns np.arrays
        tmm = TMM()
        result = tmm.fit_transform(ds.exp)
        factors = tmm.get_norm_factors(ds.exp)
        assert isinstance(result, np.ndarray)
        assert isinstance(factors, np.ndarray)

        # Explicit global config should return corresponding objects
        with config_context(transform_output="default"):
            result = tmm.fit_transform(ds.exp)
            factors = tmm.get_norm_factors(ds.exp)
            assert isinstance(result, np.ndarray)
            assert isinstance(factors, np.ndarray)

        with config_context(transform_output="pandas"):
            result = tmm.fit_transform(ds.exp)
            factors = tmm.get_norm_factors(ds.exp)
            assert isinstance(result, pd.DataFrame)
            assert isinstance(factors, pd.Series)

        # Explicit pandas config should return pandas objects
        method.set_output(transform="pandas")
        result = method.fit_transform(ds.exp)
        assert isinstance(result, pd.DataFrame)
        if has_factors:
            factors = method.get_norm_factors(ds.exp)
            assert isinstance(factors, pd.Series)



"""RNA-seq normalization methods."""

from .methods.between_sample import CTF, CUF, TMM, UQ
from .methods.utils import LibrarySize
from .methods.within_sample import CPM, FPKM, TPM

__all__ = (
    "CPM",
    "CTF",
    "CUF",
    "FPKM",
    "LibrarySize",
    "TMM",
    "TPM",
    "UQ",
)



"""GTF file parsing."""

import io
from functools import lru_cache
from pathlib import Path
from typing import Union

import pandas as pd


class GTF:
    """GTF file parser."""

    def __init__(self, gtf: Union[str, Path]) -> None:
        """Initialize class.

        :param gtf: GTF file
        """
        self.gtf = gtf

    @property
    def length(self) -> pd.Series:
        """Gene length by union exon length model."""
        return self._get_df()["gene_length"]

    @lru_cache()
    def _get_df(self) -> pd.DataFrame:
        """Load data from disc or fetch it from server and cache it on disc."""
        gtf_df = self._parse_gtf(self.gtf)

        # Compute gene length
        gene_length_df = self._gene_length(gtf_df)

        # Use only rows that represent genes
        gtf_df = gtf_df[gtf_df["feature_type"] == "gene"]
        gtf_df["gene_length"] = gene_length_df

        return gtf_df

    @staticmethod
    def _parse_gtf(file: Union[io.BytesIO, str, Path]) -> pd.DataFrame:
        """Parse GTF file."""

        # Read columns we need: 0=chromosome, 2=feature type, 3=genomic start
        # location, 4=genomic stop location, 6=genomic strand, 8=attributes
        df = pd.read_csv(
            file,
            sep="\t",
            comment="#",
            header=None,
            usecols=[0, 2, 3, 4, 6, 8],
            low_memory=False,
        )

        return pd.DataFrame(
            {
                "feature_type": df[2],
                "gene_id": df[8].str.extract(r'gene_id "(.+?)"', expand=False).values,
                "gene_name": df[8].str.extract(r'gene_name "(.+?)"', expand=False).values,
                "chromosome": df[0],
                "strand": df[6],
                "start": df[3],
                "end": df[4] + 1,  # Note: GTF is 1-based and includes stop position
            },
        ).set_index("gene_id")

    def _gene_length(self, gtf_df: pd.DataFrame, gene_id_attr: str = "gene_id") -> pd.Series:
        """Compute gene lengths based on union exon model.

        Group exon start & end coordinates by gene ID & chromosome &
        strand. Then perfrom merge and length calculation for each
        group separately. The latter is needed since ``gene_id_attr``
        is not unique in some annotations (e.g., RefSeq).
        """
        gtf_df = gtf_df[gtf_df["feature_type"] == "exon"]

        group_exons = gtf_df.groupby([gene_id_attr, "chromosome", "strand"])[["start", "end"]]
        group_length = group_exons.apply(self._compute_union_exon_length)

        gene_length = group_length.groupby(gene_id_attr).sum()

        return gene_length

    def _compute_union_exon_length(self, starts_ends: pd.DataFrame) -> pd.Series:
        """Compute gene length based on the union of its exons.

        Before summing up exon lengths, one needs to merge all
        overlapping exons.
        """
        exons = starts_ends.values
        exons = exons[exons[:, 0].argsort()]
        length = 0  # length of a rolling union of exons
        end = -1  # end of the rolling union of exons
        for exon in exons:
            if exon[1] > end:
                if exon[0] > end:
                    length += exon[1] - exon[0]
                else:
                    length += exon[1] - end
                end = exon[1]
        return length



import numpy as np
import pandas as pd
import pytest

from rnanorm import CPM


@pytest.fixture
def expected_cpm(exp):
    return pd.DataFrame(
        [
            [20000, 30000, 50000, 200000, 700000],
            [20000, 30000, 50000, 200000, 700000],
            [10000, 15000, 25000, 100000, 850000],
            [40000, 60000, 100000, 400000, 400000],
        ],
        index=exp.index,
        columns=exp.columns,
        dtype=np.float64,
    )


def test_cpm(exp, expected_cpm):
    transformer = CPM()

    # Default output is np.ndarray
    exp_normalized = transformer.fit_transform(exp)
    assert isinstance(exp_normalized, np.ndarray)
    np.testing.assert_array_equal(exp_normalized, expected_cpm.to_numpy())

    # Test nested Python list
    exp_normalized = transformer.fit_transform(exp.values.tolist())
    assert isinstance(exp_normalized, np.ndarray)
    assert exp_normalized.tolist() == expected_cpm.values.tolist()

    # Check it possible to also get pandas output
    transformer.set_output(transform="pandas")
    exp_normalized = transformer.fit_transform(exp)
    assert isinstance(exp_normalized, pd.DataFrame)
    pd.testing.assert_frame_equal(exp_normalized, expected_cpm)

    # Validation refuses nan values
    exp.iloc[0, 0] = np.nan
    with pytest.raises(ValueError, match="Input X contains NaN."):
        transformer.fit_transform(exp)



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from rnanorm import TMM
from rnanorm.datasets import load_gtex


@pytest.fixture
def expected_tmm(exp):
    return pd.DataFrame(
        [
            [20000, 30000, 50000, 200000, 700000],
            [20000, 30000, 50000, 200000, 700000],
            [20000, 30000, 50000, 200000, 1700000],
            [20000, 30000, 50000, 200000, 200000],
        ],
        index=exp.index,
        columns=exp.columns,
        dtype=np.float64,
    )


def test_tmm(exp, expected_factors, expected_tmm):
    # Simple case: Fit some data, transform same data.
    transformer = TMM()
    transformer.fit(exp)
    factors = transformer.get_norm_factors(exp)
    transformed_data = transformer.transform(exp)

    np.testing.assert_allclose(factors, expected_factors, rtol=1e-4)
    np.testing.assert_allclose(transformed_data, expected_tmm, rtol=1e-4)

    # Advanced case: Fit some samples, transform different ones.
    samples_to_fit = ["Sample_1", "Sample_3", "Sample_4"]
    exp1 = exp.loc[samples_to_fit]
    transformer.fit(exp1)
    factors = transformer.get_norm_factors(exp1)
    transformed_data = transformer.transform(exp.loc[["Sample_2"]])

    np.testing.assert_allclose(factors, [1, 0.5, 2.0], rtol=1e-4)
    np.testing.assert_allclose(
        transformed_data,
        expected_tmm.loc[["Sample_2"]],
        rtol=1e-3,
    )


def test_tmm_rnanorm_edger():
    """Test our results against EdgeR's TMM implementation."""
    # Load saved scaling factors from edgeR
    files_dir = Path(__file__).parent / "files"
    edger_factors = pd.read_csv(
        files_dir / "gtex.lung.30.chr21.tmm_factors_edgeR.tsv",
        sep="\t",
        comment="#",
        index_col=0,
        usecols=[0, 2, 3],
    )

    # Compute scaling factors here
    ds = load_gtex()
    rnanorm_factors = TMM().fit(ds.exp).get_norm_factors(ds.exp)

    # Compare loaded and computed scaling factors
    np.testing.assert_array_almost_equal(
        edger_factors["norm.factors"].values,
        rnanorm_factors,
        decimal=14,
    )



"""Between sample normalizations."""

from typing import Any, Optional

import numpy as np
import pandas as pd
from scipy.stats import gmean, rankdata, scoreatpercentile
from sklearn import config_context
from sklearn.base import BaseEstimator, OneToOneFeatureMixin, TransformerMixin
from sklearn.utils._set_output import _get_output_config
from sklearn.utils.validation import check_is_fitted, validate_data

from ..typing import Numeric1D, Numeric2D, Self
from .utils import LibrarySize


def remove_zero_genes(X: Numeric2D) -> Numeric2D:
    """Remove genes with zero count in all samples.

    Also, convert to numpy.
    """
    X = np.asarray(X)
    return X[:, np.sum(X, axis=0) > 0]


class UQ(OneToOneFeatureMixin, TransformerMixin, BaseEstimator):
    """Upper quartile (UQ) normalization.

    In an RNA-seq experiment a small fraction of genes is sometimes extremely
    overexpressed in some samples but not in others. This can artificially
    inflate library size and therefore (after library size normalization) cause
    the remaining genes to be considered under-sampled in those samples. Unless
    this effect is adjusted for, those genes may falsely appear to be
    down-regulated in that sample. Upper quartile is one of the approaches to
    correct for such imbalance. For more explanation on the topic check `EdgeR
    docs
    <https://bioconductor.org/packages/release/bioc/html/edgeR.html>`_.

    Procedure for normalization is described in `Bullard et al. 2010
    <https://bmcbioinformatics.biomedcentral.com/articles/10.1186/1471-2105-11-94>`_,
    but in short:

        - Use raw counts as input
        - Compute scaling factors
            - Remove genes that have zero count in all samples
            - Scaling factor is expression at the 75th percentile
            - Rescale factors so that their geometric mean is 1
        - "Adjusted library size" = library size * factor
        - Return CPM normalization with "Adjusted library size"

    This implementation is based on edgeR and has been validated to be
    identical to it to at least 10 decimal places.

    .. rubric:: Examples

    >>> from rnanorm.datasets import load_toy_data
    >>> from rnanorm import UQ
    >>> X = load_toy_data().exp
    >>> X
              Gene_1  Gene_2  Gene_3  Gene_4  Gene_5
    Sample_1     200     300     500    2000    7000
    Sample_2     400     600    1000    4000   14000
    Sample_3     200     300     500    2000   17000
    Sample_4     200     300     500    2000    2000
    >>> UQ().set_output(transform="pandas").fit_transform(X)
               Gene_1   Gene_2   Gene_3    Gene_4     Gene_5
    Sample_1  20000.0  30000.0  50000.0  200000.0   700000.0
    Sample_2  20000.0  30000.0  50000.0  200000.0   700000.0
    Sample_3  20000.0  30000.0  50000.0  200000.0  1700000.0
    Sample_4  20000.0  30000.0  50000.0  200000.0   200000.0
    """

    def _get_norm_factors(self, X: Numeric2D) -> Numeric1D:
        """Get UQ normalization factors (un-normalized with geometric mean).

        :param X: Expression raw count matrix (n_samples, n_features)
        """
        X = remove_zero_genes(X)
        # Make sure that global set_config(transform_output="pandas")
        # does not affect this method - we need numpy output here.
        lib_size = LibrarySize().set_output(transform="default").fit_transform(X)

        # Compute upper quartile count for each sample.
        # No numpy method can be used as drop-in replacement for R's quantile.
        # Scipy's method needs to be used, bit only works on 1D arrays
        upper_quartiles = np.apply_along_axis(
            func1d=scoreatpercentile,
            axis=1,
            arr=X,
            per=75,
        )
        return upper_quartiles / lib_size

    def get_norm_factors(self, X: Numeric2D) -> Numeric1D:
        """Get UQ normalization factors (normalized with geometric mean).

        :param X: Expression raw count matrix (n_samples, n_features)s
        """
        check_is_fitted(self)
        factors = self._get_norm_factors(X)
        factors = factors / self.geometric_mean_

        config = _get_output_config("transform", self)
        if config.get("dense", None) == "pandas" and isinstance(X, pd.DataFrame):
            return pd.Series(factors, index=X.index)
        return factors

    def _reset(self) -> None:
        """Reset internal data-dependent state."""
        if hasattr(self, "geometric_mean_"):
            del self.geometric_mean_

    def fit(self, X: Numeric2D, y: Optional[Numeric1D] = None, **fit_params: Any) -> Self:
        """Fit.

        :param X: Expression raw count matrix (n_samples, n_features)
        :return: Self
        """
        self._reset()
        X = validate_data(self, X, ensure_all_finite=True, reset=True)

        factors = self._get_norm_factors(X)
        self.geometric_mean_ = gmean(factors)

        return self

    def transform(self, X: Numeric2D) -> Numeric2D:
        """Transform.

        :param X: Expression raw count matrix (n_samples, n_features)
        :return: Normalized expression matrix (n_samples, n_features)
        """
        check_is_fitted(self)
        X = validate_data(self, X, ensure_all_finite=True, reset=False)

        # Compute effective library sizes
        factors = self.get_norm_factors(X)
        lib_size = LibrarySize().set_output(transform="default").fit_transform(X)
        effective_lib_size = lib_size * factors

        # Make CPM, but with effective library size
        return X / effective_lib_size[:, np.newaxis] * 1e6


class CUF(UQ):
    """Counts adjusted with Upper quartile factors normalization.

    Procedure for normalization is described in `Johnson & Krishnan, 2022
    <https://genomebiology.biomedcentral.com/articles/10.1186/s13059-021-02568-9>`_,
    but in short:

        - Use raw counts as input
        - Compute normalization factors the same way as in UpperQuartile
        - Divide raw counts with these factors

    .. rubric:: Examples

    >>> from rnanorm.datasets import load_toy_data
    >>> from rnanorm import CUF
    >>> X = load_toy_data().exp
    >>> X
              Gene_1  Gene_2  Gene_3  Gene_4  Gene_5
    Sample_1     200     300     500    2000    7000
    Sample_2     400     600    1000    4000   14000
    Sample_3     200     300     500    2000   17000
    Sample_4     200     300     500    2000    2000
    >>> CUF().set_output(transform="pandas").fit_transform(X)
              Gene_1  Gene_2  Gene_3  Gene_4   Gene_5
    Sample_1   200.0   300.0   500.0  2000.0   7000.0
    Sample_2   400.0   600.0  1000.0  4000.0  14000.0
    Sample_3   400.0   600.0  1000.0  4000.0  34000.0
    Sample_4   100.0   150.0   250.0  1000.0   1000.0

    """

    def transform(self, X: Numeric2D) -> Numeric2D:
        """Transform.

        :param X: Expression raw count matrix (n_samples, n_features)
        :return: Normalized expression matrix (n_samples, n_features)
        """
        check_is_fitted(self)
        X = validate_data(self, X, ensure_all_finite=True, reset=False)

        # Just divide raw counts with normalization factors
        factors = self.get_norm_factors(X)
        return X / factors[:, np.newaxis]


class TMM(OneToOneFeatureMixin, TransformerMixin, BaseEstimator):
    """Trimmed mean of M-values (TMM) normalization.

    In an RNA-seq experiment a small fraction of genes is sometimes extremely
    overexpressed in some samples but not in others . This can artificially
    inflate library size and therefore (after library size normalization) cause
    the remaining genes to be considered under-sampled in those samples. Unless
    this effect is adjusted for, those genes may falsely appear to be
    down-regulated in that sample. TMM is one of the approaches to correct for
    such imbalance. For more explanation on the topic check `EdgeR docs
    <https://bioconductor.org/packages/release/bioc/html/edgeR.html>`_.

    Procedure for normalization is described in `Robinson & Oshlack, 2010
    <https://doi.org/10.1186/gb-2010-11-3-r25>`_, but in short:

        - Use raw counts
        - Define the reference sample (``self.ref_``)
        - Compute scaling factors
            - Compute M values, filter by double trimming with m_trim
            - Compute A values, filter by double trimming with m_trim
            - Compute factors as weighted sum of M values
            - Factors = 2 ** factors
            - Rescale factors so that their geometric mean is 1
        - "Adjusted library size" = library size * normalization factors
        - Compute CPM normalization with "Adjusted library size"

    This implementation is based on edgeR's and is validated to be identical to
    it to at least 10 decimal places.

    :param m_trim: Keep genes that are within
        (``m_trim``, 1 - ``m_trim``) percentile of M-values.
    :param a_trim: Keep genes that are within
        (``a_trim``, 1 - ``a_trim``) percentile of A-values.

    .. rubric:: Examples

    >>> from rnanorm.datasets import load_toy_data
    >>> from rnanorm import TMM
    >>> X = load_toy_data().exp
    >>> X
              Gene_1  Gene_2  Gene_3  Gene_4  Gene_5
    Sample_1     200     300     500    2000    7000
    Sample_2     400     600    1000    4000   14000
    Sample_3     200     300     500    2000   17000
    Sample_4     200     300     500    2000    2000
    >>> TMM().set_output(transform="pandas").fit_transform(X)
               Gene_1   Gene_2   Gene_3    Gene_4     Gene_5
    Sample_1  20000.0  30000.0  50000.0  200000.0   700000.0
    Sample_2  20000.0  30000.0  50000.0  200000.0   700000.0
    Sample_3  20000.0  30000.0  50000.0  200000.0  1700000.0
    Sample_4  20000.0  30000.0  50000.0  200000.0   200000.0

    """

    def __init__(self, m_trim: float = 0.3, a_trim: float = 0.05) -> None:
        """Initialize class."""
        self.m_trim = m_trim
        self.a_trim = a_trim

    # XXX: Could this be cached? It is called in fit and again in transform.
    # But it is hard to cache numpy arrays...
    def _get_norm_factors(self, X: Numeric2D) -> Numeric1D:
        """Get TMM normalization factors (un-normalized with geometric mean).

        :param X: Expression raw count matrix (n_samples, n_features)
        """
        X = remove_zero_genes(X)

        # ensure that output of transform will be a np.array
        with config_context(transform_output="default"):
            lib_size = LibrarySize().fit_transform(X)
            lib_size_ref = LibrarySize().fit_transform(self.ref_[np.newaxis, :])

        # Values 0 cause a lot of troubles and warnings in log / division.
        # But computing with np.nan is OK, and is handled gracefully.
        # So convert values of 0 to np.nan early to avoid trouble.
        X[X == 0] = np.nan
        # When making 0 -> np.nan, make a copy of self.ref_, since modifying
        # self.ref_ would break library size calculation.
        ref = self.ref_.copy()
        ref[ref == 0] = np.nan

        # Equations simplify if we operate with (count / lib_size) ratio
        # Kudos for the idea to https://gitlab.com/georgy.m/conorm
        r = X / lib_size[:, np.newaxis]
        r_ref = ref / lib_size_ref

        # Gene-wise log-fold-changes, M from the paper
        # Order matters: do not change from log(x/y) to log(x)-log(y)
        m = np.log2(r / r_ref)
        # Absolute expressions levels, A from the paper
        # Order matters: do not change from log(x*y) to log(x)+log(y)
        a = np.log2(r * r_ref) / 2
        # Approximate asymptotic variances, w from the paper
        w = (1 - r) / X + (1 - r_ref) / ref

        # At this point, depart from vectorization, since removal of NaN can
        # yield different lengths of rows in m / a / w in different samples It
        # would make 2D arrays with rows of diff lengths, so not numpy
        # compatible.
        f = list()
        for i in range(X.shape[0]):
            mm = m[i]
            aa = a[i]
            ww = w[i]
            # Remove NaN values
            finite = np.isfinite(mm) & np.isfinite(aa)
            mm = mm[finite]
            aa = aa[finite]
            ww = ww[finite]

            n = len(mm)
            # Compute lower / upper cutoffs for m / a values
            m_low = np.floor(n * self.m_trim) + 1
            m_high = n - m_low + 1
            a_low = np.floor(n * self.a_trim) + 1
            a_high = n - a_low + 1

            # Determine which genes to keep
            keep = np.logical_and(
                np.logical_and(rankdata(mm) >= m_low, rankdata(mm) <= m_high),
                np.logical_and(rankdata(aa) >= a_low, rankdata(aa) <= a_high),
            )

            # Compute weighted mean of M values for genes in keep, weights w:
            # XXX: This seems to be a mistake in the paper: Paper says multiply
            # with w, but edgeR implementation divides with w. So do we.
            f.append(np.nansum(keep * mm / ww) / np.nansum(keep / ww))

        f = np.asarray(f, dtype=float)  # type: ignore[assignment]
        factors = np.power(2, f)

        return factors

    def get_norm_factors(self, X: Numeric2D) -> Numeric1D:
        """Get UQ normalization factors (normalized with geometric mean).

        :param X: Expression raw count matrix (n_samples, n_features)s
        """
        check_is_fitted(self)
        X2 = validate_data(self, X, ensure_all_finite=True, reset=False, dtype=float)

        factors = self._get_norm_factors(X2)
        factors = factors / self.geometric_mean_

        config = _get_output_config("transform", self)
        if config.get("dense", None) == "pandas" and isinstance(X, pd.DataFrame):
            return pd.Series(factors, index=X.index)
        return factors

    def _reset(self) -> None:
        """Reset internal data-dependent state."""
        if hasattr(self, "geometric_mean_"):
            del self.geometric_mean_
            del self.ref_

    def _get_ref(self, X: Numeric2D) -> Numeric1D:
        """Get reference sample."""
        f75 = UQ().fit(X).get_norm_factors(X)
        ref_index = np.argmin(np.fabs(f75 - np.mean(f75)))
        return X[ref_index, :]

    def fit(self, X: Numeric2D, y: Optional[Numeric1D] = None, **fit_params: Any) -> Self:
        """Fit.

        :param X: Expression raw count matrix (n_samples, n_features)
        :return: Self
        """
        self._reset()
        X = validate_data(self, X, ensure_all_finite=True, reset=True, dtype=float)
        X = remove_zero_genes(X)

        self.ref_ = self._get_ref(X)

        factors = self._get_norm_factors(X)
        self.geometric_mean_ = gmean(factors)

        return self

    def transform(self, X: Numeric2D) -> Numeric2D:
        """Transform.

        :param X: Expression raw count matrix (n_samples, n_features)
        :return: Normalized expression matrix (n_samples, n_features)
        """
        # Compute effective library sizes
        factors = self.get_norm_factors(X)
        if isinstance(factors, pd.Series):
            factors = factors.to_numpy()

        lib_size = LibrarySize().set_output(transform="default").fit_transform(X)
        effective_lib_size = lib_size * factors

        # Method ``check_is_fitted`` is not called here, since it is
        # called in self.get_norm_factors
        X = validate_data(self, X, ensure_all_finite=True, reset=False, dtype=float)

        # Make CPM, but with effective library size
        return X / effective_lib_size[:, np.newaxis] * 1e6


class CTF(TMM):
    """Counts adjusted with TMM factors normalization.

    Procedure for normalization is described in `Johnson & Krishnan, 2022
    <https://genomebiology.biomedcentral.com/articles/10.1186/s13059-021-02568-9>`_,
    but in short:

        - Compute normalization factors same as in TMM
        - Divide raw counts with these factors

    :param m_trim: Keep genes that are within
        (``m_trim``, 1 - ``m_trim``) percentile of M-values.
    :param a_trim: Keep genes that are within
        (``a_trim``, 1 - ``a_trim``) percentile of A-values.

    .. rubric:: Examples

    >>> from rnanorm.datasets import load_toy_data
    >>> from rnanorm import CTF
    >>> X = load_toy_data().exp
    >>> X
              Gene_1  Gene_2  Gene_3  Gene_4  Gene_5
    Sample_1     200     300     500    2000    7000
    Sample_2     400     600    1000    4000   14000
    Sample_3     200     300     500    2000   17000
    Sample_4     200     300     500    2000    2000
    >>> CTF().set_output(transform="pandas").fit_transform(X)
              Gene_1  Gene_2  Gene_3  Gene_4   Gene_5
    Sample_1   200.0   300.0   500.0  2000.0   7000.0
    Sample_2   400.0   600.0  1000.0  4000.0  14000.0
    Sample_3   400.0   600.0  1000.0  4000.0  34000.0
    Sample_4   100.0   150.0   250.0  1000.0   1000.0

    """

    def transform(self, X: Numeric2D) -> Numeric2D:
        """Transform.

        :param X: Expression raw count matrix (n_samples, n_features)
        :return: Normalized expression matrix (n_samples, n_features)
        """
        # Just divide raw counts with normalization factors
        factors = self.get_norm_factors(X)
        if isinstance(factors, pd.Series):
            factors = factors.to_numpy()

        # Method ``check_is_fitted`` is not called here, since it is
        # called in self.get_norm_factors
        X = validate_data(self, X, ensure_all_finite=True, reset=False, dtype=float)

        return X / factors[:, np.newaxis]
